#include "mpi.h"
#include "MPI_Status.h"
#include "MPI_Datatype.h"
#include "MPI_Op.h"
#include "MPI_Comm.h"

#define COMM(x) (*(MPI_Comm*)*(void**)&(x))
#define OP(x) ((MPI_Op)*(void**)&(x))
#define DATATYPE(x) ((MPI_Datatype)*(void**)&(x))
#define PTR(x) (*(void**)&(x))
#define REF(x) (*(void**)&(unhand(x)->handle))

static MPI_Comm* Comms[] = { 0, &MPI_COMM_SELF, &MPI_COMM_WORLD};

void MPI_Comm_Setup(struct HMPI_Comm *this,long Type) {
  REF(this) = Comms[Type];
}

long MPI_Comm_Barrier(struct HMPI_Comm *this) {
  return MPI_Barrier(COMM(unhand(this)->handle));
}

long MPI_Comm_Size(struct HMPI_Comm *this, 
  HArrayOfInt *Size)
{
  int size;
  int result = MPI_Comm_size(COMM(unhand(this)->handle), &size);
  *unhand(Size)->body = size;
  return result;
}

long MPI_Comm_Rank(struct HMPI_Comm *this,
  HArrayOfInt *Rank)
{
  int rank;
  int result = MPI_Comm_rank(COMM(unhand(this)->handle), &rank);
  *unhand(Rank)->body = rank;
  return result;
}  

long MPI_Comm_Send(struct HMPI_Comm *this, 
  struct HMPI_Datatype *buf,
  long dest, long tag)
{
  return MPI_Send(PTR(unhand(buf)->handle), unhand(buf)->size, 
    DATATYPE(unhand(buf)->type), dest, tag, 
    COMM(unhand(this)->handle));
}

long MPI_Comm_Bsend(struct HMPI_Comm *this, 
  struct HMPI_Datatype *buf,
  long dest, long tag)
{
  return MPI_Bsend(PTR(unhand(buf)->handle), unhand(buf)->size,
    DATATYPE(unhand(buf)->type), dest, tag, 
    COMM(unhand(this)->handle));
}

long MPI_Comm_Ssend(struct HMPI_Comm *this, 
  struct HMPI_Datatype *buf,
  long dest, long tag)
{
  return MPI_Ssend(PTR(unhand(buf)->handle), unhand(buf)->size,
    DATATYPE(unhand(buf)->type), dest, tag,
    COMM(unhand(this)->handle));
}

long MPI_Comm_Rsend(struct HMPI_Comm *this, 
  struct HMPI_Datatype *buf,
  long dest, long tag)
{
  return MPI_Rsend(PTR(unhand(buf)->handle), unhand(buf)->size,
    DATATYPE(unhand(buf)->type), dest, tag,
    COMM(unhand(this)->handle));
}

long MPI_Comm_Recv(struct HMPI_Comm *this,
  struct HMPI_Datatype *buf,
  long source, long tag,
  struct HMPI_Status *stat)
{
  MPI_Status status;
  int result = MPI_Recv(PTR(unhand(buf)->handle), unhand(buf)->size,
    DATATYPE(unhand(buf)->type), source, tag,
    COMM(unhand(this)->handle), &status);
  MPI_Get_count(&status, MPI_LONG, &status.count);
  unhand(stat)->SOURCE = status.MPI_SOURCE;
  unhand(stat)->TAG = status.MPI_TAG;
  unhand(stat)->ERROR = status.MPI_ERROR;
  unhand(stat)->count = status.count;
  return result;
}

long MPI_Comm_Bcast(struct HMPI_Comm *this,
  struct HMPI_Datatype *buf, long root)
{
  return MPI_Bcast(PTR(unhand(buf)->handle), unhand(buf)->size,
    DATATYPE(unhand(buf)->type), root, 
    COMM(unhand(this)->handle));
}

long MPI_Comm_Gather(struct HMPI_Comm *this,
  struct HMPI_Datatype *sbuf, 
  struct HMPI_Datatype *rbuf, long root)
{
  return MPI_Gather(
    PTR(unhand(sbuf)->handle), 
    unhand(sbuf)->size,
    DATATYPE(unhand(sbuf)->type),
    PTR(unhand(rbuf)->handle), 
    unhand(rbuf)->size,
    DATATYPE(unhand(rbuf)->type),
    root, COMM(unhand(this)->handle));
}

long MPI_Comm_Scatter(struct HMPI_Comm *this,
  struct HMPI_Datatype *sbuf, 
  struct HMPI_Datatype *rbuf, long root)
{
  return MPI_Gather(
    PTR(unhand(sbuf)->handle), 
    unhand(sbuf)->size,
    DATATYPE(unhand(sbuf)->type),
    PTR(unhand(rbuf)->handle), 
    unhand(rbuf)->size,
    DATATYPE(unhand(rbuf)->type),
    root, COMM(unhand(this)->handle));
}

long MPI_Comm_Allgather(struct HMPI_Comm *this,
  struct HMPI_Datatype *sbuf, 
  struct HMPI_Datatype *rbuf)
{
  return MPI_Allgather(
    PTR(unhand(sbuf)->handle), 
    unhand(sbuf)->size,
    DATATYPE(unhand(sbuf)->type),
    PTR(unhand(rbuf)->handle), 
    unhand(rbuf)->size,
    DATATYPE(unhand(rbuf)->type),
    COMM(unhand(this)->handle));
}

long MPI_Comm_Alltoall(struct HMPI_Comm *this,
  struct HMPI_Datatype *sbuf, 
  struct HMPI_Datatype *rbuf)
{
  return MPI_Alltoall(
    PTR(unhand(sbuf)->handle), 
    unhand(sbuf)->size,
    DATATYPE(unhand(sbuf)->type),
    PTR(unhand(rbuf)->handle), 
    unhand(rbuf)->size,
    DATATYPE(unhand(rbuf)->type),
    COMM(unhand(this)->handle));
}

long MPI_Comm_Reduce(struct HMPI_Comm *this,
  struct HMPI_Datatype *sbuf,
  struct HMPI_Datatype *rbuf,
  struct HMPI_Op *op, long root)
{
  return MPI_Reduce(
    PTR(unhand(sbuf)->handle), 
    PTR(unhand(rbuf)->handle), 
    unhand(sbuf)->size,
    DATATYPE(unhand(sbuf)->type),
    OP(unhand(op)->handle), root,
    COMM(unhand(this)->handle));
}

